--------------------------------------------------
-- WTI NetLink Emulation
-- (Control B commands)
--
-- Requires firmware 2.10+
--
-- 2008-10-10: Xmodem, XON dump, Multiple commands
-- 2008-10-24: Missing ready after ^B01 with ^B54,2
-- 2008-12-18: Added sanity check
-- 2009-01-20: Updated incorrect emulation date & "VER" command
-- 2009-02-23: Fixed obscure bug where record count shows 2147483647
--             and records do not download
--             Adjusted "STATUS" memory display
-- 2009-03-27: Fixed "^B01,PA" issue
-- 2009-07-20: Added support for "^B79" password changing
-- 2009-09-03: Updated ^B79 to enable parameter copying
-- 2009-10-20: Fixed ^B01 difference with ^B54,1 vs ^B54,2
-- 2014-02-07: Increased line length handling up to 8192 characters
--------------------------------------------------

if not emulation then if error then error("Needs newer firmware!",0) else Needs_Newer_Firmware() end end
c.emulation = c.emulation or {}     -- Ensure the table exists

function emulation.defaults(e)
 ---- USER PARAMETERS ----
 e._message  = "POLLCAT-NETLINK"
 e._password = "SMDR"
 e._password = (c.emulation and c.emulation.password) or e._password

 ---- DEFAULTS ----
 e._delete=false        -- Autodelete?
 e._holdend=true        -- Hold the END DATA for ^B02?
 e._suppressend=false   -- Suppress the END DATA message?
 e._echo=true           -- Echo?
 e._datafirst=true      -- Release data after ^B01?
 e._confirmdelete=true  -- Confirm deletes?
 e._compressed=false    -- Space compression?
 e._crc=false           -- CRC option
 e._binary=false        -- Whether to use binary block, or ASCII
 e._xmodem=false        -- Use xmodem in ASCII mode by default?

 e._confirmanyway=true  -- For processes that ask for Xmodem/Ymodem but use ASCII

 e._telnet=0            -- Whether to force the escape 0xff characters
 e._modem=false
 e._timeout=30000       -- Default command timeout for LAN
 if (e.peer=="127.0.0.1") -- Definitely shouldn't escape
 then
  e._telnet=-1
  e._modem=true
  e._timeout=15*60*1000 -- Large timeout for modem mode
 end
 e._shownrec=true       -- Whether to estimate number of records, or show bytes instead

 e._preventerase=false  -- Normally false
 e._endofline="\n"      -- How to split lines from memory
 e._lineid=false        -- Whether to use the line ID function
 e._disablebinarysc=false -- Disable Space Compression?
 e._disablebinarytc=false -- Disable Table Compression?

 local p=c.chnl[e.n].dst.emul.port
 e._xonrelease=(p>=2200) and (p<4000)
 //e._xonrelease=false     -- True if XON will release data
 
 e._debug=false
end


function emulation.init(e)
 e:defaults()
 e._version="2014-02-07"
 e._lineidformat="%.8d "
 e._cnt=1
 e._n=e.n
 e._linecount=0
 e._incmd=false
 e._lastsize=0
 e._lastpos=0
 e._nrec=0
 e._cmds=NULL
end

function emulation.usedbytes(e)
 local t=0
 local vi,vv
 for vi,vv in ipairs(i.mem.store)
 do t=t+vv.size
 end
 return t
end

function emulation.writeln(e,s)
 s=s or ""
 return e:write(s.."\r\n")
end

function emulation.wsf(e,f,...)
 f = f or ""
 return e:write(string.format(f.."\r\n",...))
end

function emulation.key(e,t)
 local s,t=e:read(1, t or 30000)
 if e._debug then d.client = d.client..s end
 return s,t
end

function emulation.keyline(e,echo,timeout)
 local s,t,l,sp
 l=""
 if not e._cmds
 then
  while true
  do
   s,t=e:key(timeout or e._timeout or 30000)
   if t then return NULL end
   if ((echo==NULL) and e._echo) or echo
   then
    if (s~="\x02") and (s~="\n") then e:write(s) end
    if s=="\r" then e:write("\n") end
   end
   if s=="\x08"
   then
    if ((echo==NULL) and e._echo) or echo
    then
     if l=="" then e:write(" ") else e:write(" \x08") end
    end
    l=string.sub(l,1,-2)
   elseif s=="\n"
   then
   elseif s=="\x1b"
   then return NULL
   elseif s=="\x11"
   then return "DUMP"
   elseif s=="\r"
   then
    break
   else
    l=l..s
   end
  end
 else
  l=e._cmds
  e._cmds=NULL
 end
 sp=string.find(l,"/",1,true)
 if sp
 then
  e._cmds="\x02"..string.sub(l,sp+1)
  l=string.sub(l,1,sp-1)
 end
 local cmd=string.match(l,"\x02([%a%d]*),*")  -- Control B
 if cmd
 then
  cmd=tonumber(cmd)
  l=string.sub(l, 2)
 end
 return l,cmd
end

function emulation.dataend(e)
 if not e._suppressend
 then e:writeln("\r\n\x06END DATA\r\n")
 end
end

function emulation.complete(e)
 e:writeln("COMPLETE")
end

function emulation.ready(e)
 e:writeln("READY")
end

function emulation.error(e)
 e:write("Invalid command.\r\n")
end

function emulation.compress(e,s)
 return string.gsub(s,"  *",function(c) local r="" local l=string.len(c) while l>26 do r=r.."z" l=l-26 end r=r..string.char(96+l) return r end)
end

function emulation.bytestuff(e,s)
 if e._telnet<0
 then return s
 elseif e._telnet>0 or e.telnet
 then return string.gsub(s,"\255","\255\255")
 else return s
 end
end

function emulation.word2str(e,v)
 return string.char(bit.cast(v/256),bit.band(v,255))
end

--Pull a line from the memory (complex!)
function emulation.getline(e)
 local s,n=e:get(250)
 if n<=0 then return "",0,true end
 local _,t=string.find(s,e._endofline)
 if t
 then
  e:adv(t)
  s=string.sub(s,1,t)
  return s,string.len(s),false
 else
  e:adv(n)
  while string.len(s) < 8192
  do
    local s2,n2=e:get(2048)
    _,t=string.find(s2,e._endofline)
    if not t
    then
      e:adv(n2)
      s=s..s2
    else
      e:adv(t)
      s2=string.sub(s2,1,t)
      s=s..s2
      return s,string.len(s),false
    end
  end
  return s.."\r\n",string.len(s),false
 end
end

--Count and estimate how many records
function emulation.estnrec(e)
 if not e._shownrec
 then
  e._nrec=e:frozen()
  return e._nrec
 end
 e._lastsize=0
 e._nrec=0
 e:reset(0)
 e:adv(0)
 if e:frozen()<=0 then return e._nrec end
 while ((e._lastsize<8000) and (e._nrec<200))
 do
  local s,ll=e:getline()
  e._lastsize=e._lastsize+ll
  if s=="" then break end
  e._nrec=e._nrec+1
 end
 e._nrec=math.floor((e._nrec/e._lastsize)*e:frozen()+0.5)
 e:reset()
 e:adv(0)
 e._lastsize=0
 return e._nrec
end

--Output lines in ASCII mode
function emulation.outputlines(e)
 local l=e._cnt
 local ds=false
 local c=crc.crcxmodem_init()
 e._lastsize=0
 while ((l<0) or (l>0)) and not e:dead()
 do
  local s,ll=e:getline()
  e._lastsize=e._lastsize+ll
  if s=="" then break end
  if e._lineid
  then
   e._linecount=e._linecount+1
   s=string.format(e._lineidformat,e._linecount)..s
  end
  ds=true
  if e._compressed then s=e:compress(s) end
  if e._crc then c=crc.crcxmodem_update(c,s) end
  s=e:bytestuff(s)
  e:write(s)
  l=l-1
 end
 if e._crc
 then
  c=crc.crcxmodem_finish(c)
  e:write(e:bytestuff(e:word2str(c)))
 end
 if (not e._holdend or not ds or (l<0)) then e:dataend() return false end
 return true
end

--Binary Block Space Compression
function emulation._bbsc(e,s)
 if string.find(s,"[\128-\255]") then return s,false end
 return string.gsub(s,"   +",
  function(c)
   local r=""
   local l=string.len(c)
   while l>127 do r=r.."\255";l=l-127;end
   r=r..string.char(128+l)
   return r
  end),true
end

--Binary Block Table Compression
function emulation._bbtc(e,s)
 local t="0123456789 :/\r\n"
 local r=string.gsub(s,"(.)",
  function(c)
   local p=string.find(t,c,1,true)
   if p
   then return string.format("%X",p-1)
   else
    p=string.byte(c)
    return "F"..string.format("%X%X",(p%16),math.floor(p/16))
   end
  end)
 r=r.."0"
 r=string.gsub(r,"(%x%x)",
  function(c)
   return string.char(tonumber(c:reverse(),16))
  end)
 return r
end

--Form and output block, wait for reply
function emulation.bb_block(e,s,count,ending,noreply)
 if ending
 then s="\255"..e:word2str(count)..s
 else s="\000"..e:word2str(count)..s
 end
 e._linecount=e._linecount+1
 if e._lineid then s=e:word2str(e._linecount)..s end
 s="\002"..e:word2str(string.len(s))..s.."\003"
 local c=crc.crcxmodem(s)
 s=s..e:word2str(c)
 local tries=0
 while tries<2
 do
  e:write(e:bytestuff(s))
  if noreply then return false end
  local l,cmd=e:keyline(e._echo, 5000)
  if cmd==02
  then return true
  elseif cmd==30
  then if ending then return false else e:bb_enddata(true) return false end
  end
  tries=tries+1
 end
 if not ending then e:bb_enddata(true) end
 return false
end

--Form a binary record
function emulation.bb_rec(e,s,nsc,ntc)
 local m=0
 local res
 local r
 local lo=string.len(s)
 if lo>252 then lo=252 s=string.sub(s,1,lo) end
 nsc=nsc or e._disablebinarysc
 ntc=ntc or e._disablebinarytc
 if not nsc then s,res=e:_bbsc(s) if res then m=m+2 end end
 if not ntc then r=e:_bbtc(s) if string.len(r)<string.len(s) then s=r m=m+4 end end
 return string.char(string.len(s)+3,m,lo)..s
end

--Output END DATA in binary
function emulation.bb_enddata(e,noreply)
 e._ended=source.timestamp("%m/%d/%y %H:%M\r\n\r\n\004")
 local s=e:bb_rec(string.format("END DATA  %s-%s",e._started,e._ended),true,true)
 return e:bb_block(s,1,true,noreply)
end

--Output records in binary
function emulation.bb_output(e,first)
 local o=""
 local nrec=0
 local more=true
 e._lastsize=0
 while string.len(o)<400
 do
  local s,ll=e:getline()
  e._lastsize=e._lastsize+ll
  if s=="" then more=false break end
  o=o..e:bb_rec(s)
  nrec=nrec+1
 end
 if (nrec==0) and first then e:bb_enddata(true) return false,false end
 return e:bb_block(o,nrec,false),more
end

function emulation.releasebinary(e,cmd)
 local more,res,l,cmd
 e:reset()
 e:adv(0)
 e._lastsize=0
 e._started=source.timestamp("%m/%d/%y %H:%M")
 if not e._datafirst
 then
  e:ready()
  while true
  do
   l,cmd=e:keyline()
   if cmd==30 then e:bb_enddata(true) return
   elseif cmd==02 then break
   end
  end
 end
 local first=true
 while true
 do
  res,more=e:bb_output(first)
  first=false
  if not res then return end
  if not e._preventerase and e._delete
  then
   e._lastsize=0
   e._lastpos=0
   e:erase()
  end
  if not more then e:bb_enddata(false) return end
 end
end

function emulation.releaseascii(e,cmd)
 local n=string.match(cmd,",(%d+),*")
 if n then e._cnt=tonumber(n) else e._cnt=-1 end
 e:reset()
 e:adv(0)
 e._lastsize=0
 if e._datafirst
 then
  if not e:outputlines() then return end
 else
  e:ready()
 end
 while true
 do
  local l,cmd=e:keyline()
  if cmd==30 then e:dataend() return
  elseif cmd==02
  then
   if e._delete and not e._preventerase
   then
    e._lastsize=0
    e._lastpos=0
    e:erase()
   else
    e._lastpos=e._lastpos+e._lastsize
   end
   if not e:outputlines() then return end
  elseif cmd==06
  then
   e:reset()
   e:adv(e._lastpos)
   if not e:outputlines() then return end
  end
 end
end

function emulation._releasexmodem(e,cmd)
 local seq=1
 local pkt=""
 local load=true
 local out=true
 local usecrc=false
 e:reset()
 e:adv(0)
 while true
 do
  if e:dead() then return true end
  local s,t=e:key(30000)
  if t then return true end
  if s=="\x18" then return true end
  if string.find(s,"[C\x06\x15]") 
  then 
   if string.find(s,"[C]")
   then usecrc=true
   end
   break
  end
 end
 while true
 do
  if load
  then
   local s=""
   local l=0
   while (l<128)
   do
    local ss,ll=e:get(128-l)
    if (ll<=0)
    then
     if string.len(s)==0 then return false end
     s=s..string.rep("\x00",128-l)
     break
    end
    e:adv(ll)
    s=s..ss
    l=l+ll
   end
   if usecrc
   then
    chk=crc.crcxmodem(s)
    pkt="\x01"..string.char(seq, 255-seq)..s..string.char((chk/256)%256)..string.char(chk%256)
   else
    local chk=0
    local idx=1
    while idx<=128
    do
     chk=chk+string.byte(s,idx)
     idx=idx+1
    end
    pkt="\x01"..string.char(seq, 255-seq)..s..string.char(chk%256)
   end
   seq=seq+1
   if seq>255 then seq=0 end
   load=false
  end

  if out
  then
   e:write(e:bytestuff(pkt))
   out=false
  end

  if e:dead() then return true end
  local s,t=e:key(30000)
  if t then return end
  if s=="\x18" then return true
  elseif s=="\x06" then
   load=true
   out=true
  elseif s=="\x15" then
   out=true
  end
 end
end

function emulation.releasexmodem(e,cmd)
 if e:_releasexmodem(cmd) then return false end
 while true
 do
  if e:dead() then return false end
  e:write("\x04")
  local s,t=e:key(30000)
  if t then return end
  if s=="\x06"
  then
   if e._delete
   then e:erase()
   end
   return
  end
 end
end

function emulation.dumpdata(e)
 if not e._xonrelease
 then
  e:error()
  return
 end
 local pause=false
 while not e:dead()
 do
  while true
  do
   k=e:read(1)
   if not k or (k=="") then break end
   if k=="\x13" then pause=true end
   if k=="\x11" then pause=false end
   if k=="\x18" then return end
   if e:dead() then return end
  end
  while pause
  do
   k,t=e:key()
   if t then return false end
   if k=="\x11" then pause=false end
  end
  local s,n=e:get(250)
  if n<=0
  then
   e:dataend()
   e:erase()
   return true
  end
  e:adv(n)
  if not e:write(e:bytestuff(s)) then return false end
 end
end

function emulation.b01_releasemode(e,cmd)
 e._linecount=0
 if not e._binary
 then
  if e._xmodem
  then return emulation.releasexmodem(e,cmd)
  else return emulation.releaseascii(e,cmd)
  end
 else return emulation.releasebinary(e,cmd)
 end
end

function emulation.openmem(e,show)
 if e._n>0
 then
  e._lastsize=0
  e._lastpos=0
  if not e:open(e._n) then e:write("Could not open file (in use?)!") return end
  e:freeze()
  if show
  then
   if (e:frozen()<=0)
   then e:writeln("No data available.")
   else e:complete()
   end
  end
 end
end

function emulation.b00_memorypartition(e,cmd)
 e._n=0
 if string.find(cmd, "R") then e:close() e:complete()
 elseif string.find(cmd, "PA") then e._n=1
 elseif string.find(cmd, "PB") then e._n=2
 elseif string.find(cmd, "PX") then e._n=3
 elseif string.find(cmd, "P1") then e._n=1
 elseif string.find(cmd, "P2") then e._n=2
 elseif string.find(cmd, "P3") then e._n=3
 elseif string.find(cmd, "P4") then e._n=4
 else e._n=e.n
 end
 e:openmem(true)
 e:estnrec()
end

function emulation.b03_siteid(e,cmd)
 if e._n>0
 then e:wsf("Site ID Msg: %s.%s", c.network.name, c.chnl[e._n].name)
 else e:wsf("Site ID Msg: %s", c.network.name)
 end
end

function emulation.howmany(e,v)
 if v>(1024*1024) then return string.format("%8d Mbytes",v/(1024*1024)+0.5)
 elseif v>1024 then return string.format("%8d kbytes",v/1024+0.5)
 else return string.format("%8d bytes",v)
 end
end

function emulation.howmanyshort(e,v)
 if v>(1024*1024) then return string.format("%d MB",v/(1024*1024)+0.5)
 elseif v>1024 then return string.format("%d KB",v/1024+0.5)
 else return string.format("%d B",v)
 end
end

function emulation.b20_bytes(e,cmd)
 if (e._n > 0)
 then
  if e._shownrec
  then e:wsf("Partition Rec Ct: %8d", e._nrec)
  else e:wsf("Partition Rec Ct: %s", e:howmany(i.mem.store[e._n].size))
  end
 end
end

function emulation.b21_free(e,cmd)
 e:wsf("Space Remaining: %s",e:howmany(i.mem.blocks.erased*i.mem.blocks.size))
end

function emulation.b59_lineid(e,cmd)
 if string.find(cmd,",2")
 then e._lineid=true
 else e._lineid=false
 end
 e:complete()
end

function emulation.b67_autodel(e,cmd)
 if string.find(cmd,",2")
 then e._delete=true
 else e._delete=false
 end
 e:complete()
end

function emulation.b09_quit(e,cmd)
 e:writeln("Command mode exited.")
 return 1
end

function emulation.times(e,tv)
 if not tv or (tv==0) then return "N/A" end
 return os.date("%Y-%m-%d %H:%M:%S",tv)
end

function emulation.full_status(e,cmd)
 local n=e._n
 e:wsf("Scannex ip.buffer %s",c.firmware)
 e:wsf("Emulation version %s",e._version)
 e:wsf("%s  %3d%% full", e:times(i.seconds), (e:usedbytes() * 100 / (i.mem.blocks.totalstore * i.mem.blocks.size)))
 e:writeln()
 if (n<=0) then e:writeln("No channel selected.") return end
 e:wsf("#%d '%s'", n, c.chnl[n].name)
 if i.chnl[n].src.connected==1
 then e:writeln("Source: Connected")
 else e:writeln("Source: !DISCONNECTED!")
 end
 local st=i.chnl[n].src.type
 local si=i.chnl[n].src
 local di=i.chnl[n].dst
 local ci=c.chnl[n].src
 local qt=si.data.quiet
 e:wsf(" Quiet: %d-%.2d:%.2d:%.2d",qt/(24*60*60),(qt/(60*60))%24,(qt/60)%60,qt%60)
 if st=="tcp"
 then
  e:wsf("   TCP: peer=%s", si.tcp.peer)
  if si.tcp.connected==1
  then
   if (si.tcp.sendmatch>=4)
   then e:writeln("        Connected!")
   else e:wsf("        Connected & matching %d",si.tcp.sendmatch)
   end
  else e:writeln("        Not connected!")
  end
 elseif st=="com"
 then
  e:wsf("   COM: rxpin=%s baud=%s,%s", si.com.rxpin,ci.com.baud,si.com.protocol)
  e:wsf("        rts=%d dtr=%d cts=%d dsr=%d", si.com.rts,si.com.dtr,si.com.cts,si.com.dsr)
  if (si.com.autobauding==1) then e:writeln("        Autobauding!") end
 elseif st=="ftps"
 then e:wsf("   FTP: peer=%s last=%d bytes", si.ftps.peer, si.ftps.lastsize)
 elseif st=="udp"
 then e:wsf("   UDP: ???")
 else e:wsf("UNKNOWN SOURCE %s!", st)
 end
 e:writeln()
 e:writeln("STORAGE");
 e:wsf(" Bytes: %8d", i.mem.store[n].size)
 e:wsf("  Lost: %8d new, %8d old", i.mem.store[n].lostnew, i.mem.store[n].lostold)
 e:writeln()
 e:writeln("DESTINATION");
 if not di.last.start
 then
  e:writeln(" Never connected!")
 else
  local inf=di.last.remote

  if di.last.intf=="M" then inf=inf.." (modem)" end
  if di.last.ssl=="M" then inf=inf.." SSL" end
  e:wsf("  Last: %s", inf or "")
  e:wsf(" Start: %s", e:times(di.last.start))
  e:wsf("   End: %s", e:times(di.last["end"]))
  if di.last.frozen and (di.last.frozen>0)
  then e:wsf("  Sent: %8d of %8d bytes", di.last.transferred or 0, di.last.frozen or 0)
  else e:wsf("  Sent: %8d bytes", di.last.transferred or 0)
  end
  e:wsf(" Error: %s", di.last.error or "")
 end
 e:writeln()
 e:complete()
end

function emulation.b16_status(e,cmd)
 local perc=math.floor(((e:usedbytes()*100)/(i.mem.blocks.totalstore*i.mem.blocks.size))+0.5)
 local ismdm=(e.peer=="127.0.0.1")
 e:wsf("\r\nScannex ip.buffer   Version %11s %s  %d%% Full", e._version, source.timestamp("%m/%d/%Y %a %H:%M"),perc)
 e:wsf("BUFFER STATUS:\r\n")
 e:wsf("Unit Capacity:      %s",e:howmanyshort(i.mem.blocks.totalstore*i.mem.blocks.size))
 e:wsf("Space Used:         %s",e:howmanyshort(i.mem.store[e._n].size))
 e:wsf("Space Remaining:    %s",e:howmanyshort(i.mem.blocks.erased*i.mem.blocks.size))
 e:wsf("Percent Full:       %d%%",perc)
 e:wsf("Total Record Ct:    %d", e._nrec)
 e:wsf("\r\nPARTITION RECORD COUNT")
 e:wsf("Console Port:       None")
 if ismdm
 then e:wsf("Modem Port:         %d",e._nrec)
 else e:wsf("Modem Port:         None")
 end
 if not ismdm
 then e:wsf("Telnet Port #1:     %d",e._nrec)
 else e:wsf("Telnet Port #1:     None")
 end
 e:wsf("Telnet Port #2:     None")
 e:wsf("Telnet Port #3:     None")
 e:wsf("FTP Push:           None")
 e:wsf("FTP Server:         None")
 e:wsf("\r\n\r\n\r\n\r\n\r\n")
end

function emulation.b25_erasepart(e,cmd)
 if not string.find(cmd, ",[Yy]") and e._confirmdelete
 then
  e:writeln("WARNING: All partition data will be erased.\r\nSure ? (Y/N):")
  local l=e:keyline()
  if not l or not string.find(l,"[Yy]") then e:writeln("ABORT") return end
 end
 if e._n<=0 then e:writeln("No partition has been set by this port.") return end
 if e._preventerase then e:complete() return end
 if not e:erasefrozen()
 then e:wsf("!ERROR ERASING CHANNEL %d", e._n)
 else e:complete()
 end
 e:openmem(false)
 e:estnrec()
end

function emulation.b52_enddata(e,cmd)
 if string.find(cmd,",1")   -- v3.27 works this way (opposite to all the manuals!!)
 then e._holdend=true
 else e._holdend=false
 end
 e:complete()
end

function emulation.b54_datafirst(e,cmd)
 if string.find(cmd,",1")
 then e._datafirst=true
 else e._datafirst=false
 end
 e:complete()
end

function emulation.b62_format(e,cmd)
 if not string.find(cmd,",[1234]") and not e._confirmanyway
 then
  e:writeln("Invalid command. !ONLY SUPPORTS ASCII LINES, BINARY BLOCK & XMODEM!")
  return
 end
 if string.find(cmd,",3")
 then e._binary=true
 else
  e._binary=false
  if string.find(cmd,",4")
  then e._xmodem=true
  else e._xmodem=false
  end
 end
 e:complete()
end

function emulation.b68_echo(e,cmd)
 if string.find(cmd,",2")
 then e._echo=true
 else e._echo=false
 end
 e:complete()
end

function emulation.b72_date(e,cmd)
 e:writeln(i.now)
end

function emulation.bxx_sayok(e,cmd)
 e:complete()
end

function emulation.b75_sure(e,cmd)
 if string.find(cmd,",1")
 then e._confirmdelete=false
 else e._confirmdelete=true
 end
 e:complete()
end

function emulation.b53_compress(e,cmd)
 if string.find(cmd,",1")
 then e._compressed=true
 else e._compressed=false
 end
 e:complete()
end

function emulation.b63_crc(e,cmd)
 if string.find(cmd,",2")
 then e._crc=true
 else e._crc=false
 end
 e:complete()
end

function emulation.b71_dle(e,cmd)
 if string.find(cmd,",1")
 then e:complete()
 else e:writeln("Only ^B71,1 supported")
 end
end

function emulation.b79_password(e,cmd)
 local pw=string.match(cmd, ",([^,]+)$")
 if not pw
 then e:writeln(e._password)
 else
  e._password = pw
  c.emulation = c.emulation or {}
  c.emulation.password = pw
  _changed = true
  e:complete()
 end
end


--------------------------------------
-- Lua mode handling
--------------------------------------
function emulation.wluavar(e,name,val)
 if type(val)=="string"
 then e:wsf("%s = %q",name,val)
 elseif type(val)=="number"
 then e:wsf("%s = %s",name,val)
 elseif type(val)~="function"
 then e:wsf("%s = [Unknown]",name)
 end
end

function emulation.luadump(e,root,tbl)
 local key,val,full
 for key,val in pairs(tbl)
 do
  if string.sub(key,1,1)~="_"
  then
   if type(key)=="number"
   then full=string.format("%s[%s]",root,key)
   else full=string.format("%s.%s",root,key)
   end
   if type(val)=="table"
   then e:luadump(full,val)
   else e:wluavar(full,val)
   end
  end
 end
end

function emulation.luamode(e)
 local l,n,v,cur,uc
 e:write("Username:")
 n=e:keyline(false)
 if not n then return false end
 e:writeln()
 e:write("Password:")
 v=e:keyline(false)
 if not v then return false end
 if (n~=c.web.setup.user) or (v~=c.web.setup.pass)
 then e:writeln("Invalid username/password!") return true
 else e:writeln()
 end
 while not e:dead()
 do
  e:write("Lua>")
  l=e:keyline(true,5*60*1000)
  uc=string.upper(l or "")
  if not l or (uc=="BYE") or (uc=="QUIT")
  then
   e:writeln("\r\nExiting Lua mode")
   return true
  end
  n=string.match(l,"%s*([%a%d%p]*)%s*=")
  if not n
  then n=l v=NULL
  else v=string.match(l,"=([%a%d%p%s]*)")
  end
  cur=NULL
  if n then cur=tools.get(n) end
  if not cur then e:writeln("Non existent variable")
  else
   if v
   then
    v=string.gsub(v,"[\"']","")
    if type(cur)=="string" or type(cur)=="number"
    then tools.set(n,tools.unquote(v)) cur=tools.get(n) e:wluavar(n,cur)
    else e:writeln("Incomplete variable name")
    end
   else
    if type(cur)=="string" or type(cur)=="number"
    then e:wluavar(n,cur)
    else e:luadump(n,cur)
    end
   end
  end
 end
end

--------------------------------------
-- Command mode
--------------------------------------
emulation.cmds = {
  //[02]=emulation.cmdnext, -- Only in datamode
  //[06]=emulation.cmdresend, -- Only in datamode
  //[30]=emulation.cmdexitcmd, -- Only in datamode
  [00]=emulation.b00_memorypartition,
  [01]=emulation.b01_releasemode,
  [03]=emulation.b03_siteid,
  [09]=emulation.b09_quit,
  [11]=emulation.bxx_sayok, --unsupported
  [16]=emulation.b16_status,
  [20]=emulation.b20_bytes,
  [21]=emulation.b21_free,
  [25]=emulation.b25_erasepart,
  [52]=emulation.b52_enddata,
  [53]=emulation.b53_compress,
  [54]=emulation.b54_datafirst,
  [59]=emulation.b59_lineid,
  [62]=emulation.b62_format,
  [63]=emulation.b63_crc,
  [67]=emulation.b67_autodel,
  [68]=emulation.b68_echo,
  [71]=emulation.b71_dle,
  [72]=emulation.b72_date,
  [74]=emulation.bxx_sayok, --not applicable
  [75]=emulation.b75_sure,
  [79]=emulation.b79_password
  }

function emulation.signon(e)
 local pw=""
 e:writeln(e._message)
 while true
 do
  pw=e:keyline(false)
  if not pw then e:writeln("\r\nPassword timeout.") return false end
  if pw==e._password then return true end
  e:writeln("ERROR")
  e:writeln(e._message)
 end
end

function emulation.commandmode(e)
 while true
 do
  local l,cmd=e:keyline()
  if not l then e:writeln("\r\nInactivity timeout.") return false end
  if cmd
  then
   local fn=emulation.cmds[cmd]
   if not fn
   then e:error()
   else return not fn(e,string.sub(l,2))
   end
  else
   if l==""
   then e:write("Enter selection ... ") return true
   end
   local _,r,cmd=string.find(l,"(%a+)%s*")
   if (r) then l=string.sub(l,r+1) end
   cmd=string.upper(cmd or "")
   if cmd=="LUA"
   then
    if not e:luamode() then return false end
    e:complete()
   elseif cmd=="STATUS"
   then e:full_status()
   elseif cmd=="BYE" or cmd=="QUIT"
   then return not e:b09_quit()
   elseif cmd=="FREE"
   then e:b21_free()
   elseif cmd=="COUNT"
   then e:b20_bytes()
   elseif cmd=="ZERO"
   then e:b25_erasepart()
   elseif cmd=="VER"
   then
    e:wsf("Scannex ip.buffer")
    e:wsf("   Serial %s",i.device.serial)
    e:wsf(" Firmware %s",c.firmware)
    e:wsf("Emulation %s",e._version)
    e:wsf("COM Ports %d",i.device.serialports)
    e:wsf(" Channels %d",i.device.channels)
    e:wsf("    Files %d",i.device.files)
    e:wsf("   Memory %s",e:howmanyshort(i.mem.blocks.totalstore*i.mem.blocks.size))
    e:wsf("  Battery %d",i.system.batteryfitted)
    if (i.modem.country~="")
    then e:wsf("    Modem Fitted")
    else e:wsf("    Modem not fitted")
    end
    e:wsf("     Time %s",i.now)
   elseif cmd=="DUMP"
   then
    e:dumpdata()
    e:sleep(5)
    return false
   end
  end
 end
end

function emulation.connect(e)
 e:init()
 if e._debug
 then
  d.client_last = d.client
  d.client=""
 end
 if e._modem then e:sleep(2) end
 if not e:signon() then return end
 e:openmem(false)
 e:estnrec()
 e:ready()
 while e:commandmode()
 do --
 end
end
